@@ -111,11 +111,11 @@ getbitval(
 			/* handle endian swap here */
 #if __BYTE_ORDER == LITTLE_ENDIAN
 			if (i == 0 && signext && nbits < 64)
-				rval = -1LL << nbits;
+				rval = (~0ULL) << nbits;
 			rval |= 1ULL << (nbits - i - 1);
 #else
 			if ((i == (nbits - 1)) && signext && nbits < 64)
-				rval |= (-1LL << nbits);
+				rval |= ((~0ULL) << nbits);
 			rval |= 1ULL << (nbits - i - 1);
 #endif
 		}
