@@ -525,6 +525,7 @@ _("data fork in rt inode %" PRIu64 " found metadata block %" PRIu64 " in rt bmap
 		case XR_E_INUSE:
 			if (pwe)
 				break;
+			/* fall through */
 		case XR_E_MULT:
 			set_rtbmap(ext, XR_E_MULT);
 			do_warn(
@@ -750,7 +751,6 @@ _("%s fork in ino %" PRIu64 " claims free block %" PRIu64 "\n"),
 				/* fall through ... */
 			case XR_E_INUSE1:	/* seen by rmap */
 			case XR_E_UNKNOWN:
-				set_bmap_ext(agno, agbno, blen, XR_E_INUSE);
 				break;
 
 			case XR_E_BAD_STATE:
@@ -772,7 +772,6 @@ _("%s fork in inode %" PRIu64 " claims metadata block %" PRIu64 "\n"),
 
 			case XR_E_INUSE:
 			case XR_E_MULT:
-				set_bmap_ext(agno, agbno, blen, XR_E_MULT);
 				if (type == XR_INO_DATA &&
 				    xfs_sb_version_hasreflink(&mp->m_sb))
 					break;
@@ -791,6 +790,34 @@ _("%s fork in %s inode %" PRIu64 " claims CoW block %" PRIu64 "\n"),
 				do_error(
 _("illegal state %d in block map %" PRIu64 "\n"),
 					state, b);
+				goto done;
+			}
+		}
+
+		/*
+		 * Update the internal extent map only after we've checked
+		 * every block in this extent.  The first time we reject this
+		 * data fork we'll try to rebuild the bmbt from rmap data.
+		 * After a successful rebuild we'll try this scan again.
+		 * (If the rebuild fails we won't come back here.)
+		 */
+		agbno = XFS_FSB_TO_AGBNO(mp, irec.br_startblock);
+		ebno = agbno + irec.br_blockcount;
+		for (; agbno < ebno; agbno += blen) {
+			state = get_bmap_ext(agno, agbno, ebno, &blen);
+			switch (state)  {
+			case XR_E_FREE:
+			case XR_E_FREE1:
+			case XR_E_INUSE1:
+			case XR_E_UNKNOWN:
+				set_bmap_ext(agno, agbno, blen, XR_E_INUSE);
+				break;
+			case XR_E_INUSE:
+			case XR_E_MULT:
+				set_bmap_ext(agno, agbno, blen, XR_E_MULT);
+				break;
+			default:
+				break;
 			}
 		}
 		if (collect_rmaps) { /* && !check_dups */
@@ -1276,6 +1303,119 @@ null_check(char *name, int length)
 	return(0);
 }
 
+/*
+ * This does /not/ do quotacheck, it validates the basic quota
+ * inode metadata, checksums, etc.
+ */
+#define uuid_equal(s,d) (platform_uuid_compare((s),(d)) == 0)
+static int
+process_quota_inode(
+	struct xfs_mount	*mp,
+	xfs_ino_t		lino,
+	struct xfs_dinode	*dino,
+	uint			ino_type,
+	struct blkmap		*blkmap)
+{
+	xfs_fsblock_t		fsbno;
+	struct xfs_buf		*bp;
+	xfs_filblks_t		dqchunklen;
+	uint			dqperchunk;
+	int			quota_type;
+	char			*quota_string;
+	xfs_dqid_t		dqid;
+	xfs_fileoff_t		qbno;
+	int			i;
+	int			t = 0;
+
+	switch (ino_type) {
+		case XR_INO_UQUOTA:
+			quota_type = XFS_DQ_USER;
+			quota_string = _("User quota");
+			break;
+		case XR_INO_GQUOTA:
+			quota_type = XFS_DQ_GROUP;
+			quota_string = _("Group quota");
+			break;
+		case XR_INO_PQUOTA:
+			quota_type = XFS_DQ_PROJ;
+			quota_string = _("Project quota");
+			break;
+		default:
+			ASSERT(0);
+	}
+
+	dqchunklen = XFS_FSB_TO_BB(mp, XFS_DQUOT_CLUSTER_SIZE_FSB);
+	dqperchunk = libxfs_calc_dquots_per_chunk(dqchunklen);
+	dqid = 0;
+	qbno = NULLFILEOFF;
+
+	while ((qbno = blkmap_next_off(blkmap, qbno, &t)) != NULLFILEOFF) {
+		xfs_dqblk_t	*dqb;
+		int		writebuf = 0;
+
+		fsbno = blkmap_get(blkmap, qbno);
+		dqid = (xfs_dqid_t)qbno * dqperchunk;
+
+		bp = libxfs_readbuf(mp->m_dev, XFS_FSB_TO_DADDR(mp, fsbno),
+				    dqchunklen, 0, &xfs_dquot_buf_ops);
+		if (!bp) {
+			do_warn(
+_("cannot read inode %" PRIu64 ", file block %" PRIu64 ", disk block %" PRIu64 "\n"),
+				lino, qbno, fsbno);
+			return 1;
+		}
+
+		dqb = bp->b_addr;
+		for (i = 0; i < dqperchunk; i++, dqid++, dqb++) {
+			int		bad_dqb = 0;
+
+			/* We only print the first problem we find */
+			if (xfs_sb_version_hascrc(&mp->m_sb)) {
+				if (!libxfs_verify_cksum((char *)dqb,
+							sizeof(*dqb),
+							XFS_DQUOT_CRC_OFF)) {
+					do_warn(_("%s: bad CRC for id %u. "),
+							quota_string, dqid);
+					bad_dqb = 1;
+					goto bad;
+				}
+
+				if (!uuid_equal(&dqb->dd_uuid,
+						&mp->m_sb.sb_meta_uuid)) {
+					do_warn(_("%s: bad UUID for id %u. "),
+							quota_string, dqid);
+					bad_dqb = 1;
+					goto bad;
+				}
+			}
+			if (libxfs_dquot_verify(mp, &dqb->dd_diskdq, dqid,
+						quota_type, 0) != NULL) {
+				do_warn(_("%s: Corrupt quota for id %u. "),
+						quota_string, dqid);
+				bad_dqb = 1;
+			}
+
+bad:
+			if (bad_dqb) {
+				if (no_modify)
+					do_warn(_("Would correct.\n"));
+				else {
+					do_warn(_("Corrected.\n"));
+					libxfs_dquot_repair(mp, &dqb->dd_diskdq,
+							    dqid, quota_type);
+					writebuf = 1;
+				}
+			}
+		}
+
+		if (writebuf && !no_modify)
+			libxfs_writebuf(bp, 0);
+		else
+			libxfs_putbuf(bp);
+	}
+	return 0;
+}
+
 static int
 process_symlink_remote(
 	struct xfs_mount	*mp,
@@ -1329,6 +1469,13 @@ _("cannot read inode %" PRIu64 ", file block %d, disk block %" PRIu64 "\n"),
 				lino, i, fsbno);
 			return 1;
 		}
+		if (bp->b_error == -EFSCORRUPTED) {
+			do_warn(
+_("Corrupt symlink remote block %" PRIu64 ", inode %" PRIu64 ".\n"),
+				fsbno, lino);
+			libxfs_putbuf(bp);
+			return 1;
+		}
 		if (bp->b_error == -EFSBADCRC) {
 			do_warn(
 _("Bad symlink buffer CRC, block %" PRIu64 ", inode %" PRIu64 ".\n"
@@ -1478,6 +1625,13 @@ _("size of socket inode %" PRIu64 " != 0 (%" PRId64 " bytes)\n"), lino,
 _("size of fifo inode %" PRIu64 " != 0 (%" PRId64 " bytes)\n"), lino,
 				(int64_t)be64_to_cpu(dino->di_size));
 			break;
+		case XR_INO_UQUOTA:
+		case XR_INO_GQUOTA:
+		case XR_INO_PQUOTA:
+			do_warn(
+_("size of quota inode %" PRIu64 " != 0 (%" PRId64 " bytes)\n"), lino,
+				(int64_t)be64_to_cpu(dino->di_size));
+			break;
 		default:
 			do_warn(_("Internal error - process_misc_ino_types, "
 				  "illegal type %d\n"), type);
@@ -1598,7 +1752,7 @@ process_check_sb_inodes(
 	int		*dirty)
 {
 	if (lino == mp->m_sb.sb_rootino) {
-	 	if (*type != XR_INO_DIR)  {
+		if (*type != XR_INO_DIR)  {
 			do_warn(_("root inode %" PRIu64 " has bad type 0x%x\n"),
 				lino, dinode_fmt(dinoc));
 			*type = XR_INO_DIR;
@@ -1612,7 +1766,7 @@ process_check_sb_inodes(
 		return 0;
 	}
 	if (lino == mp->m_sb.sb_uquotino)  {
-		if (*type != XR_INO_DATA)  {
+		if (*type != XR_INO_UQUOTA)  {
 			do_warn(_("user quota inode %" PRIu64 " has bad type 0x%x\n"),
 				lino, dinode_fmt(dinoc));
 			mp->m_sb.sb_uquotino = NULLFSINO;
@@ -1621,7 +1775,7 @@ process_check_sb_inodes(
 		return 0;
 	}
 	if (lino == mp->m_sb.sb_gquotino)  {
-		if (*type != XR_INO_DATA)  {
+		if (*type != XR_INO_GQUOTA)  {
 			do_warn(_("group quota inode %" PRIu64 " has bad type 0x%x\n"),
 				lino, dinode_fmt(dinoc));
 			mp->m_sb.sb_gquotino = NULLFSINO;
@@ -1630,7 +1784,7 @@ process_check_sb_inodes(
 		return 0;
 	}
 	if (lino == mp->m_sb.sb_pquotino)  {
-		if (*type != XR_INO_DATA)  {
+		if (*type != XR_INO_PQUOTA)  {
 			do_warn(_("project quota inode %" PRIu64 " has bad type 0x%x\n"),
 				lino, dinode_fmt(dinoc));
 			mp->m_sb.sb_pquotino = NULLFSINO;
@@ -1737,6 +1891,14 @@ _("directory inode %" PRIu64 " has bad size %" PRId64 "\n"),
 			return 1;
 		break;
 
+	case XR_INO_UQUOTA:
+	case XR_INO_GQUOTA:
+	case XR_INO_PQUOTA:
+		/* Quota inodes have same restrictions as above types */
+		if (process_misc_ino_types(mp, dino, lino, type))
+			return 1;
+		break;
+
 	case XR_INO_RTDATA:
 		/*
 		 * if we have no realtime blocks, any inode claiming
@@ -2092,14 +2254,9 @@ process_inode_attr_fork(
 		do_warn(_("bad attribute fork in inode %" PRIu64), lino);
 
 		if (!no_modify)  {
-			if (delete_attr_ok)  {
-				do_warn(_(", clearing attr fork\n"));
-				*dirty += clear_dinode_attr(mp, dino, lino);
-				dino->di_aformat = XFS_DINODE_FMT_LOCAL;
-			} else  {
-				do_warn("\n");
-				*dirty += clear_dinode(mp, dino, lino);
-			}
+			do_warn(_(", clearing attr fork\n"));
+			*dirty += clear_dinode_attr(mp, dino, lino);
+			dino->di_aformat = XFS_DINODE_FMT_LOCAL;
 			ASSERT(*dirty > 0);
 		} else  {
 			do_warn(_(", would clear attr fork\n"));
@@ -2110,7 +2267,7 @@ process_inode_attr_fork(
 		blkmap_free(ablkmap);
 		*retval = 1;
 
-		return delete_attr_ok ? 0 : 1;
+		return 0;
 	}
 
 	if (check_dups)  {
@@ -2204,6 +2361,28 @@ _("would clear obsolete nlink field in version 2 inode %" PRIu64 ", currently %d
 	return dirty;
 }
 
+/* Check nanoseconds of a timestamp don't exceed 1 second. */
+static void
+check_nsec(
+	const char		*name,
+	xfs_ino_t		lino,
+	struct xfs_timestamp	*t,
+	int			*dirty)
+{
+	if (be32_to_cpu(t->t_nsec) < 1000000000)
+		return;
+
+	do_warn(
+_("Bad %s nsec %u on inode %" PRIu64 ", "), name, be32_to_cpu(t->t_nsec), lino);
+	if (no_modify) {
+		do_warn(_("would reset to zero\n"));
+	} else {
+		do_warn(_("resetting to zero\n"));
+		t->t_nsec = 0;
+		*dirty = 1;
+	}
+}
+
 /*
  * returns 0 if the inode is ok, 1 if the inode is corrupt
  * check_dups can be set to 1 *only* when called by the
@@ -2330,7 +2509,8 @@ process_dinode_int(xfs_mount_t *mp,
 			if (!uncertain)
 				do_warn(
 _("inode identifier %llu mismatch on inode %" PRIu64 "\n"),
-					be64_to_cpu(dino->di_ino), lino);
+					(unsigned long long)be64_to_cpu(dino->di_ino),
+					lino);
 			if (verify_mode)
 				return 1;
 			goto clear_bad_out;
@@ -2509,6 +2689,18 @@ _("bad (negative) size %" PRId64 " on inode %" PRIu64 "\n"),
 			flags2 &= XFS_DIFLAG2_ANY;
 		}
 
+		if (flags2 & XFS_DIFLAG2_DAX) {
+			/* must be a file or dir */
+			if (di_mode && !(S_ISREG(di_mode) || S_ISDIR(di_mode))) {
+				if (!uncertain) {
+					do_warn(
+	_("DAX flag set on special inode %" PRIu64 "\n"),
+						lino);
+				}
+				flags2 &= ~XFS_DIFLAG2_DAX;
+			}
+		}
+
 		if ((flags2 & XFS_DIFLAG2_REFLINK) &&
 		    !xfs_sb_version_hasreflink(&mp->m_sb)) {
 			if (!uncertain) {
@@ -2611,6 +2803,12 @@ _("bad (negative) size %" PRId64 " on inode %" PRIu64 "\n"),
 			type = XR_INO_RTBITMAP;
 		else if (lino == mp->m_sb.sb_rsumino)
 			type = XR_INO_RTSUM;
+		else if (lino == mp->m_sb.sb_uquotino)
+			type = XR_INO_UQUOTA;
+		else if (lino == mp->m_sb.sb_gquotino)
+			type = XR_INO_GQUOTA;
+		else if (lino == mp->m_sb.sb_pquotino)
+			type = XR_INO_PQUOTA;
 		else
 			type = XR_INO_DATA;
 		break;
@@ -2681,6 +2879,7 @@ _("Cannot have non-zero CoW extent size %u on non-cowextsize inode %" PRIu64 ",
 					be32_to_cpu(dino->di_cowextsize), lino);
 			if (!no_modify)  {
 				do_warn(_("resetting to zero\n"));
+				dino->di_flags2 &= ~cpu_to_be64(XFS_DIFLAG2_COWEXTSIZE);
 				dino->di_cowextsize = 0;
 				*dirty = 1;
 			} else
@@ -2688,6 +2887,31 @@ _("Cannot have non-zero CoW extent size %u on non-cowextsize inode %" PRIu64 ",
 		}
 	}
 
+	/*
+	 * Can't have the COWEXTSIZE flag set with no hint.
+	 */
+	if (dino->di_version >= 3 &&
+	    be32_to_cpu(dino->di_cowextsize) == 0 &&
+	    (be64_to_cpu(dino->di_flags2) & XFS_DIFLAG2_COWEXTSIZE)) {
+		do_warn(
+_("Cannot have CoW extent size of zero on cowextsize inode %" PRIu64 ", "),
+				lino);
+		if (!no_modify)  {
+			do_warn(_("clearing cowextsize flag\n"));
+			dino->di_flags2 &= ~cpu_to_be64(XFS_DIFLAG2_COWEXTSIZE);
+			*dirty = 1;
+		} else {
+			do_warn(_("would clear cowextsize flag\n"));
+		}
+	}
+
+	/* nsec fields cannot be larger than 1 billion */
+	check_nsec("atime", lino, &dino->di_atime, dirty);
+	check_nsec("mtime", lino, &dino->di_mtime, dirty);
+	check_nsec("ctime", lino, &dino->di_ctime, dirty);
+	if (dino->di_version >= 3)
+		check_nsec("crtime", lino, &dino->di_crtime, dirty);
+
 	/*
 	 * general size/consistency checks:
 	 */
@@ -2756,6 +2980,15 @@ _("Cannot have non-zero CoW extent size %u on non-cowextsize inode %" PRIu64 ",
 			goto clear_bad_out;
 		}
 		break;
+	case XR_INO_UQUOTA:
+	case XR_INO_GQUOTA:
+	case XR_INO_PQUOTA:
+		if (process_quota_inode(mp, lino, dino, type, dblkmap) != 0) {
+			do_warn(
+	_("problem with quota inode %" PRIu64 "\n"), lino);
+			goto clear_bad_out;
+		}
+		break;
 	default:
 		break;
 	}
